/* $Id: servsock.c,v 1.67 2004/04/21 19:54:02 geschke Exp $ */
/* $Revision: 1.67 $ */
/*
** This program is part of the FLoP project
**
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/**************************************************************************/
/*                                                                        */
/* Main program to receive snort data via TCP and send them through unix  */
/* socket to a mysql or postgres database. To avoid blocking during read/ */
/* write we use synchronised POSIX threads. The parent process listens    */
/* for incoming requests and forks for every sensor. Statistics are       */
/* gathered through the use of signals. A SIGUSR1 send to the parent      */
/* results in sending a SIGUSR2 signal to all child process which in      */
/* reaction print the statistics to stdout or to syslog with facility     */
/* LOCAL0 and level INFO                                                  */
/* A further feature is writing of alerts based on the priority to a unix */
/* domain socket                                                          */
/*                                                                        */
/* Author: Dirk Geschke                                                   */
/*                                                                        */
/**************************************************************************/

static char *rcsid="$Id: servsock.c,v 1.67 2004/04/21 19:54:02 geschke Exp $";

#include "FLoPconfig.h"
#undef PACKAGE
#undef VERSION
#undef PACKAGE
#undef VERSION
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#include "FLoP_endian.h"
#include "socket_database.h"
#include <sys/types.h>
#include <stdio.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <pthread.h>
#include "spo_alert_unixsock_db.h"
#include <unistd.h>
#include <signal.h>
#include "Signal.h"
#include "comm.h"
#include "io.h"

/* the new getipnodebyaddr and getipnodebyname are (at least) on linux */
/* only in libbind (aka named) available (maybe we should use the old  */
/* IPv4 versions instead?                                              */
#if defined HAVE_BIND_HEADER && defined LINUX
#include <bind/netdb.h>
#else
#include <netdb.h>
#endif

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX   108
#endif

#ifdef GENUA
#warning GENUA defined!
#endif

/* if DEBUG is defined, define every debug option */
#ifdef DEBUG
#define DEBUG_MAIN
#define DEBUG_WTS
#define DEBUG_RFC
#define DEBUG_CMDLINE
#define DEBUG_OS
#define DEBUG_OAS
#define DEBUG_ODS
#define DEBUG_READCONFIG
#define DEBUG_CHLD
#define DEBUG_DB
#define DEBUG_DROPPACKETS
#endif


#define MAXCHILD 25   /* max number of child processes we observe */
/* Note: even if there are more then MAXCHILD child processes the */
/*       program will work, we only loose the statistics of the   */
/*       not registered child processes. To be correct we have to */
/*       use a linked list and malloc memory for each new process */
/*       but this can lead to problems in connection with signals */

#define MINWATER 100
#define MINWATERDIFF 10

#if defined ALERT || defined DROP
#include "alert.h"
#endif

/* TYPEDEFS */

typedef struct _DGPacket 
{
  u_int8_t *pkt_ptr;            /* pointer to the network packet */
  AlertpktDB alertpkt;          /* AlterpktDB from snort...     */
  struct _DGPacket *next;       /* pointer to the next packet  */
  struct _DGPacket *prev;       /* pointer to the previous packet  */
} DGPacket;


typedef struct _RemoteSensor
{
  pid_t pid;
  u_int32_t ip;
} RemoteSensor;

                     /* PROTOTYPES */

int OpenServer(char *servername, int port);
void *ReceiveFromClient(void *ptr);
void *WriteToSocket(void *ptr);
void Usage(char *argv);
void CmdLine(int argc, char *argv[]);
void SigChld();
void SigUsr1();
void SigUsr2();
void ReadConfig();
void DumpConfig();
void SigAlarm();
void SigInt();
void SigInt2();
void SigHup();
void ArgCheck(char *arg);
char *Time();
int CheckIP(u_int32_t ip);
int ReadData(char *SwapFile);
int DumpData(char * SwapFile,DGPacket* data);
int getdaemon(char *dir, mode_t mask);

#ifdef ALERT
void OpenAlertSocket(char *filename);
int send2socket(void * data, int size);
#endif

#ifdef DROP
void SigDrop();
void OpenDropSocket(char *filename);
int drop2socket(void * data, int size);
void DropPackets(DGPacket *last);
#endif

sighandler_t Signal(int signr, sighandler_t sighandler);



/* Global data */
pthread_mutex_t mutex;   /* mutex to block and send signals on new data */
pthread_cond_t thr_cond;   
DGPacket *data;
DGPacket *StartDBdata;
static u_int ServerPort;
static u_int SocketNameSet=0;
static u_int ServerNameSet=0;
static u_int ServerPortSet=0;
static u_int AlarmDelaySet=0;
static u_int DBtrustSet=0;
static u_int DBtransSet=0;
static u_int FQNSensorSet=0;
static u_int FQNSensor=1;
u_int LogSyslog=0;
u_int Daemon=0;
int FullPayload=0;
u_int DaemonSet=0;
u_int DaemonDirSet=0;
u_int SwapDirSet=0;
u_int maskSet=0;
u_int PIDFileSet=0;
mode_t mask=0;
int DBFailed=0;

#ifdef ALERT
/* global file descriptor for unix domain socket */
static int UnixSocket=0;
static int UnixSocketSet=0;
static char UnixSocketName[STRLEN]="/tmp/alert";
struct sockaddr_un alert_addr;
int UnixSocketPriority=2;
int UnixSocketPrioritySet=0;
#endif

#ifdef DROP
static int DropSocket=0;
static int DropSocketSet=0;
static char DropSocketName[STRLEN]="/tmp/drop";
struct sockaddr_un drop_addr;
u_int HighWater;
int HighWaterSet=0;
u_int LowWater;
int LowWaterSet=0;
#endif
pthread_t RFCthread=0;
pthread_t WTSthread=0;
u_int DropCounter=0;

pthread_t MainThread=0;

static char SocketName[STRLEN]="NULL";
static char ServerName[STRLEN]="0.0.0.0";
static char ConfigFile[STRLEN]="servsock.conf";
static char DaemonDir[STRLEN]="";
static char SwapDir[STRLEN]="/var/tmp";
static char SwapFile[STRLEN]="";
char PIDFile[STRLEN]="/tmp/servsock.pid";
static char DBuser[STRLEN]="snort";
static char DBpassword[STRLEN]="";
static char DBencoding[STRLEN]="hex";
static char DBname[STRLEN]="snort";
static int DBtrust=0;
static int DBtrans=0;
static int DBtype=0;
static RemoteSensor Child[MAXCHILD];
DGPacket *WTSdata;


/* counter used within the threads to gather statistics. They are */
/* used with the signal handler so we have to define them global  */
static u_int WTScounter=0;
static u_int RFCcounter=0;
static u_int RSTcounter=0;

/* used to for a clean exit if client closes unexpectedly the connetion */
/* this is needed to fill in the (maybe) last data to the database      */
u_int clientexit=0; 
u_int DumpConf=0; 

u_int AlarmDelay=30;

DatabaseData *DBdata;

char Endian;

/* needed for restart */
char *prog;
char **args;
int ServerSocket;

int 
main(int argc, char* argv[])
{
  pthread_t thr_wts, thr_rfc;
  pthread_attr_t thr_attr;
  int sock1,conn;
  int clientlength;
  char line[STRLEN];
  char errbuf[STRLEN];
  FILE *fd;
#ifdef DEBUG_MAIN
  int clientport;
#endif
  struct sockaddr_in clientin;
  pid_t pid;
  int storepid=0;
  int i,j;
  int ret;
  DGPacket *previous;

  /* A simple hack to decompose IP4 addresses in dotted parts */
  union { 
    struct {u_int8_t b1, b2, b3, b4; } IPb;
    u_int32_t IPc;
  } IP;
  
  struct hostent *hostinfo;
  char SensorName[STRLEN];

  /* Store progname and arguments for restart */
  prog=argv[0];
  args=argv;

  snprintf(ConfigFile,STRLEN,"%s/servsock.conf",PATH_CONFIG);

  for (i=0; i<MAXCHILD; i++) 
    {
      Child[i].pid=0;
      Child[i].ip=0;
    }

  bzero((void *)&hostinfo,sizeof(hostinfo));

  /* Check endian of this host */
  Endian=CheckEndian();

  /* To store the configuration of the database */
  DBdata=(DatabaseData *)calloc(1, sizeof(DatabaseData));
  if (DBdata == NULL) 
    {
      Printf("MAIN> ERROR: Unable to allocate memory!\n");
      exit(1);
    }

  pid=getpid();
  /* install signal handler to recognize end of child process */
  Signal(SIGCHLD,SigChld);  
  Signal(SIGUSR1,SigUsr1);
  Signal(SIGINT,SigInt);
  Signal(SIGTERM,SigInt);
  Signal(SIGHUP,SigHup);  
  /* Main process should ignore USR2, this is for the clients only */
  Signal(SIGUSR2,SIG_IGN);

  /* Parse command line arguments, maybe we get a new config file... */
  CmdLine(argc,argv);

  /* Read config file */
  ReadConfig();

  /* Check availability of the swap directory */
  if (CheckDirectory(SwapDir))
    {
      Printf("No access to swap directory. We need 'rwx' access rights!\n");
      exit(1);
    }
  /* Dump configuration we have. Note: command line arguments */
  /* have precedence above settings in config file!           */
  if (DumpConf)  DumpConfig();

  /* Set the umask */
  umask(mask);

  if (Daemon) 
    {
      /* Switch into daemon mode */
      Printf("starting in daemon mode.\n");
      if ( DaemonDir[0]=='\0') 
      {
         if (getcwd(DaemonDir,STRLEN)==NULL)
	 { 
	   DG_strerror_r(errno,errbuf,STRLEN);
           Printf("can't get working directory: %s\n",errbuf);
           exit(1);
	  }
      }
      if (getdaemon(DaemonDir,mask))
        {
	  DG_strerror_r(errno,errbuf,STRLEN);
          Printf("requested daemon mode failed: %s\n",errbuf);
	  exit(1);
        }
      /* Since daemon mode closes all file descriptors we
       * close the syslog and reopen it to avoid writing to
       * the first opened filedescriptor which would be the
       * connecting sockserv process
       */
      closelog();
      openlog(prog,LOG_PID,LOG_LOCAL0);
    }

  /* Check for existing PID file and if there is a running process      */
  /* with this pid. We don't check which process runs, only if one does */
  if ( (fd=fopen(PIDFile,"r")) ) 
    {
      if (fgets(line,STRLEN,fd))
	{
	  pid=(pid_t) atoi(line);
	  if ( !kill(pid,0) )
	    {
	      Printf("Program seems to be running with pid: %u\n",(u_int) pid);
	      Printf("if not: remove pid file: %s\n",PIDFile);
	      fclose(fd);
	      exit(1);
	    }
	}
      fclose(fd);
    }
  
  /* get actual pid */
  pid=getpid();

  /* Try to write the pid to the PIDFile, removing any old one */
  if ( (fd=fopen(PIDFile,"w")) )
    {
      fprintf(fd,"%u\n",(u_int) pid);
      fclose(fd);
    }
  else
    {
      /* We are not able to write the pid file */
      Printf("Can not write pid file: %s",PIDFile);
      exit(1); /* Maybe we simply should ignore this? */
    }


#ifdef ALERT
  if (UnixSocket) OpenAlertSocket(UnixSocketName);
#endif

#ifdef DROP
  if (DropSocket)
    {
      if (LowWater < MINWATER) 
	{ 
	  LowWater=MINWATER;
	  Printf("LowWater too small -> increasing to %i\n",LowWater);
	}
      if ( (HighWater-LowWater) < MINWATERDIFF )
	{
	  Printf("Water mark difference too low -> disabling drop feature!\n");
	  Printf("Increase difference to at least %i\n",MINWATERDIFF);
	  DropSocket=0;
	}
    }
  if (DropSocket) OpenDropSocket(DropSocketName);
#endif

  if (!LogSyslog) printf("\nMain PID: %i\n",pid);

  /* if AlarmDelay is set: install timer routine to display */
  /* statistics about collected and delivered alerts with a */
  /* time interval of AlarmDelay seconds                    */
  if (AlarmDelay) 
    {
      Signal(SIGALRM,SigAlarm);
      alarm(AlarmDelay);
    }

  if (DBtype==0)
    {
      Printf("Wrong or no DBtype, allowed are:\n");
#ifdef MYSQL_DB
	      Printf("MySQL\n");
#endif
#ifdef POSTGRES
	      Printf("Postgres\n");
#endif
	      exit(1);
    }
	      

  DBdata->encoding=-1;
  if(!strncasecmp(DBencoding,"hex",3)) DBdata->encoding = ENCODING_HEX; 
  if(!strncasecmp(DBencoding,"base64",6)) DBdata->encoding = ENCODING_BASE64;
  if(!strncasecmp(DBencoding, "ascii", 5)) DBdata->encoding = ENCODING_ASCII;
  if (DBdata->encoding==-1)
     {
         Printf("\nUnknown database encoding: %s\n",DBencoding);       
	 exit(1);
     }

  DBdata->password    = DBpassword;
  DBdata->dbname      = DBname;
  DBdata->user        = DBuser;
  if (!strncmp(SocketName,"NULL",4)) SocketName[0]='\0';
  DBdata->socket_name = SocketName;
  DBdata->DBtrust     = DBtrust;
  DBdata->transaction = DBtrans;
  DBdata->DBtype      = DBtype;
  DBdata->FullPayload = FullPayload;
  /* DBdata->sensor_name must be set after client has connected! */

  /* pointer to global structure */
  /* this is the common start point for both threads */
  data = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
  if (data == NULL) 
    {
      Printf("MAIN> ERROR: Unable to allocate memory!\n");
      exit(1);
    }
  /* 
   * DB starts at data too. But if we re-read data from a file
   * we change the data pointer for the RFC thread but still need
   * the old start point to begin with database thread
   */
  StartDBdata = data;
#ifdef DEBUG_MAIN
  Printf("MAIN> Data at %p\n",data);
#endif

  /* Initialize threads */
  pthread_mutex_init(&mutex,NULL);
  pthread_attr_init(&thr_attr);

  /* should be joinable by default but who knows?    */
  /* This way we can wait for the finish of a thread */
  /* and react on it, close connections and restart  */
  /* the server or wait for new incoming request     */ 
  pthread_attr_setdetachstate(&thr_attr, PTHREAD_CREATE_JOINABLE);

  /* Store thread id for main thread */
  MainThread=pthread_self();

  /* Open TCP socket for server process */
  sock1=OpenServer(ServerName,ServerPort);
  ServerSocket=sock1;

  clientlength = sizeof(clientin);

  /* start an endless loop in server listening mode */
  while(1) 
    {
      conn = accept(sock1,(struct sockaddr*)&clientin,&clientlength);
      if (conn == -1 ) 
	{
          if (errno==EINTR) continue;
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Accept error: %s\n",errbuf);
	  exit(1);
	}
      if (CheckIP(clientin.sin_addr.s_addr)) 
	{
	  snprintf(errbuf,STRLEN,"300 We have still data to be inserted");
          WriteLn(conn,errbuf);
	  close(conn);
	  continue;
	}
      /* find an empty slot to store the pid (this is done in child process) */
      for (i=0; i<MAXCHILD; i++)
	{
	  if (Child[i].pid==0) 
	    {
	      /*
	       * store temporarily a value to avoid race condition if
	       * two client connects use the same index i 
	       * This value is replaced in the parent process
               * with the real process id after fork.
	       */
	      storepid=i;
	      Child[i].pid=1;
	      break;
	    }
	}
      if (i==MAXCHILD)
	{
	  Printf("Too many sensors connected!\n");
	  Printf("Increase MAXCHILD and recompile.\n");
	  snprintf(errbuf,STRLEN,"300 Too many childs connected.");
          WriteLn(conn,errbuf);
	  close(conn);
	  /*	  exit(1);  */
	}
      if ((pid=fork())<0)
	{
	  /* fork() failed, all we can do is to exit... */
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Fork error: %s\n",errbuf);
	  exit(1);
	}
      else if (pid > 0)
	{
	  /* Parent: close connection and wait for next */
          /* incoming request of an client/sensor       */
#ifdef DEBUG_MAIN
	  Printf("Main> Child created: %u\n",pid);
#endif 
	  /* store pid */
	  Child[storepid].pid=pid;
	  Child[storepid].ip=clientin.sin_addr.s_addr;
	  close(conn);
	  continue;
	}
      else 
	{
	  /* Child: exit loop via break and continue */
	  close(sock1); 
	  break;
	}
    }

  /* Here only child processes */

  Signal(SIGALRM,SIG_IGN);
  Signal(SIGUSR1,SIG_IGN);
  Signal(SIGUSR2,SigUsr2);
  Signal(SIGHUP,SIG_IGN);
  Signal(SIGINT,SigInt2);
  Signal(SIGTERM,SigInt2);

#ifdef DEBUG_MAIN
  /* Port of client, not really needed... */
  clientport=ntohs(clientin.sin_port);
  Printf("MAIN> Clientport %i\n",clientport);
#endif

  /* we use the new IP version independent functions  */
  /* but these are not implementet on all systems yet */
  /* you can find them in libbind (www.isc.org)       */
#ifdef HAVE_BIND_HEADER
  errno=0;
  hostinfo = getipnodebyaddr((const void * )&clientin.sin_addr,
                               sizeof(struct in_addr),AF_INET,&errno);
#else  
   hostinfo = gethostbyaddr((const void * )&clientin.sin_addr,
                            sizeof(struct in_addr),AF_INET);
#endif

  /* get IP address in network byte order */
  IP.IPc= clientin.sin_addr.s_addr;

#if defined ALERT || defined DROP
  /* store sensor address in numerical format */
  DBdata->sensorIP=IP.IPc;
#endif

  /* Fill in IP Address in dotted format as sensor_name to use */
  /* with the database. Of course a full qualified name would  */
  /* be better. But what we have... */
  snprintf(SensorName, STRLEN, "%u.%u.%u.%u", 
		  IP.IPb.b1, IP.IPb.b2, IP.IPb.b3, IP.IPb.b4);

  /* filename for the case we have to swap out data 
   * we take the ip address to omit possible problems with DNS
   */
  snprintf(SwapFile, STRLEN,  "%s/sensor_%s",SwapDir,SensorName);

  if ( hostinfo == NULL )
    /* we were not able to get the fqn */
    Printf("MAIN> No fqn: %i\n",errno);
  else
   {
     if (FQNSensor) 
	{
           /* overvrite IP address by fqn */
           snprintf(SensorName, STRLEN, "%s", hostinfo->h_name);
        } 
   }

  DBdata->sensor_name=SensorName;
#ifdef DEBUG_MAIN
  Printf("MAIN> Clientname: %s\n",DBdata->sensor_name);
#endif
  
  Printf("Sensor: %s has connected, Child: %i\n",DBdata->sensor_name,getpid());

  /* finally free the allocted memory for DNS lookup */
#ifdef HAVE_BIND_HEADER
  freehostent(hostinfo); 
#endif

  /* Connect to database, initialize some structures */
  ret=DatabaseInit(DBdata);
  if (ret)
    {
      if (ret == 1)
	snprintf(errbuf,STRLEN,"300 Problem with database, try later");
      else
	snprintf(errbuf,STRLEN,"400 Serious database error");
      Printf("Database Problem: <%s>\n",errbuf);
      WriteLn(conn,errbuf);
      close(conn);
      exit(0);
    }
  ret=CheckFile(SwapFile);
  if (ret<0) 
    {
      /* An error occured -> close connection and exit */
      snprintf(errbuf,STRLEN,"400 Serious error with swap file");
      Printf("CheckFile problem: <%s>\n",errbuf);
      WriteLn(conn,errbuf);
      close(conn);
      exit(0);
    }
  else if (ret>0)
    {
      /* We have a swap file with data */
      j=ReadData(SwapFile);
      if (j < 0)
	{
	  /* An error occured */
	  snprintf(errbuf,STRLEN,"%s_bad",SwapFile);
	  if (rename(SwapFile,errbuf))
	    {
	      DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("Error renaming swapfile %s_bad: %s\n",
		     SwapFile, errbuf);
	      Printf("Removing %s\n",SwapFile);
	      remove(SwapFile);
	    }
	  else
	    {
	      snprintf(errbuf,STRLEN,"Serious error with swap file, moved" 
		       " it to %s_bad",SwapFile);
	      Printf("ReadData problem: <%s>\n",errbuf);
	    }
	  /* Now we free allocated data */
	  while((previous=data->prev) != NULL)
	    {
	      if(data->pkt_ptr) free(data->pkt_ptr);
	      free(data);
	      data=previous;
	    }
	  
	  /* Reset old startpoint (should be equal?) */
	  data=StartDBdata;
	}
      else
	{
	  Printf("Restored %i alerts for sensor %s\n",j,SensorName);
	  /* Set RSTcounter to restored number */
	  RSTcounter=j;
	  /* We do not need the swap file so we remove it */
	  remove(SwapFile);
	}
    }

  snprintf(errbuf,STRLEN,"200 Endian Check: <%i>",Endian);
  WriteLn(conn,errbuf);
  /* Read endianess */
  ReadLn(conn,errbuf,STRLEN);
  if ((Endian+200) != atoi(errbuf))
    {
      /* Endian mismatch. Here we cancel the session now */
      snprintf(errbuf,STRLEN,"500 Error: Endian mismatch");
      WriteLn(conn,errbuf);
      Disconnect(DBdata);
      close(conn);
      exit(1);
    }
  else
    {
      snprintf(errbuf,STRLEN,"200 OK: Endian matches");
      if (WriteLn(conn,errbuf)==-1)
        {
           Printf("Error writing answer: Exiting\n");
           close(conn);
           exit(0);
        }
    }
  
  /* start first thread: Feeding to database via unix socket */
  errno=pthread_create(&thr_wts,NULL,WriteToSocket,(void *)DBdata);
  if (errno != 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error with WriteToSocket thread: %s\n",errbuf);
      exit(1);
    }

  /* start second thread: Read input from client via TCP */
  errno=pthread_create(&thr_rfc,NULL,ReceiveFromClient,(void *)&conn);
  if (errno != 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error with ReceiveFromClient thread: %s\n",errbuf);
      exit(1);
    }

  /* threads attributes no longer needed */
  pthread_attr_destroy(&thr_attr);

  /* Wait for end of threads. Should never happen? */
  /* But maybe one process get stopped, network    */
  /* problems may arise or simply someone kills    */
  /* the client program, the box crashes,...       */

  pthread_join(thr_rfc,NULL);
#ifdef DEBUG_MAIN
  Printf("MAIN> Exit of ReceiveFromClient thread!\n");
#endif

  pthread_join(thr_wts,NULL);
#ifdef DEBUG_MAIN
  Printf("MAIN> Exit of WriteToSocket thread!\n");
#endif

  /* Both threads are finished now: Do a cleanup and exit */
  pthread_mutex_destroy(&mutex);
  
  /* Disconnect database and free allocated memory */
  Disconnect(DBdata);
  if (DBdata) free(DBdata);

  /* do the exit... */
  pthread_exit(NULL);

  /* Don't ask. It is useless here, but if we close this socket*/ 
  /* before we call pthread_exit the cpu usage raises and the  */
  /* process will not exit!!! (at least on a linux system...)  */
  close(sock1);
}


/*
 * Function: WriteToSocket
 *
 * Purpose: Thread to send the snort data to a (mysql) database
 *
 * Arguments: Pointer to the database information like socket
 *
 * Author: DG
 */

void *
WriteToSocket(void *ptr) {

  DatabaseData *DBdata_ptr;
  DGPacket *tmp;
#ifdef DEBUG_WTS 
  char* text;
  time_t time;
#endif
  AlertpktDB *p_ptr;
  void *pkt_ptr;
  Event *event_ptr;
#ifdef ALERT
  AlertData alert;
#endif
  int success;

#ifdef DEBUG_WTS  
  Printf("Start WTS\n");
#endif

  /* Store the information regarding the database: */
  /* socket, user, dbname, password,...            */
  DBdata_ptr=(DatabaseData *) ptr; 

  /* Starting point of data on first entry of thread */
  WTSdata=StartDBdata;

  /* Store actual thread */
  WTSthread=pthread_self();

  /* Start endless loop: Storing data to database via socket */  
  while(1){
#ifdef DEBUG_WTS
    Printf("\nWTS> data      at %p\n",WTSdata);
    Printf("WTS> data next at %p\n",WTSdata->next);
#endif
    if (WTSdata->next == NULL) 
      {
	/* No data available, so wait for incoming packets */
        /* The thread ReceiveFromClient will send a signal */
        /* if a new packet is available.                   */
#ifdef DEBUG_WTS
	Printf("WTS> Waiting for mutex %u\n",WTScounter);
#endif

	/* I think this is not needed here but a few lines */
        /* below. But it is no error to have it twice...   */
        if (clientexit)
	  {
	    if (DBFailed)
	      Printf("DB failure, we closed connection, total entries: %i\n",
		     WTScounter);
	    else
	      Printf("Sensor: %s has closed connection, total entries: %i\n",
		     DBdata_ptr->sensor_name,WTScounter);
	    
	    pthread_exit(NULL);
	  }

	pthread_mutex_lock(&mutex);
	pthread_cond_wait(&thr_cond,&mutex);
	pthread_mutex_unlock(&mutex);

        /* if clientexit is set then we have no new data to */
        /* send to the database, the other thread will not  */
        /* get any more data. So all we can do here is to   */
        /* exit the thread. This hack is needed to keep the */
        /* thread WTS as long running as we have data which */
        /* needs to be send to the database                 */
        if (clientexit)
	  {
	    if (DBFailed)
	      Printf("DB failure, we closed connection, total entries: %i\n",
                      DBdata_ptr->sensor_name,WTScounter);
	    else
	      Printf("Sensor: %s has closed connection, total entries: %i\n",
                      DBdata_ptr->sensor_name,WTScounter);
	    
	    pthread_exit(NULL);
	  }
      }

#ifdef DEBUG_WTS      
      time=(time_t) WTSdata->alertpkt.pkth.ts.tv_sec;
      if (time>0 ) 
        {
          Printf("WTS> Time %i: %s",(int) time,ctime(&time));
        }
      else 
	{
          Printf("WTS> No time!\n");
	}
#ifndef GENUA
      Printf("WTS> Alertmsg: %s\n",WTSdata->alertpkt.alertmsg);
#endif
      Printf("WTS> Event ID : %i\n",WTSdata->alertpkt.event.event_id);
      Printf("WTS> Priority : %i\n",WTSdata->alertpkt.event.priority);
#endif

      /* Initalize pointer to payload with NULL, maybe there is */
      /* no payload available...                                */
      pkt_ptr=NULL;
      if(!(WTSdata->alertpkt.val & NOPACKET_STRUCT)) 
	{

#ifdef DEBUG_WTS
	  text=Fasthex((u_char *)WTSdata->pkt_ptr,
		       WTSdata->alertpkt.pkth.caplen);
	  Printf("WTS> Got mutex %u\n",WTScounter);
	  Printf("WTS> Paketsize: %u\n",WTSdata->alertpkt.pkth.caplen);
	  Printf("WTS> Content  : %s\n",text);
#endif

	  pkt_ptr=(void *)WTSdata->pkt_ptr;
	}
     
      /* Pointer to the alertpkt */
      p_ptr=&(WTSdata->alertpkt);
   
      /* pointer to the event structure located in alertpkt */
      event_ptr=&WTSdata->alertpkt.event;

      /* Write the data to the database */
#ifdef GENUA
      success = Database(p_ptr, NULL, DBdata_ptr, event_ptr, pkt_ptr);
#else
      success = Database(p_ptr, (char *)WTSdata->alertpkt.alertmsg, 
			 DBdata_ptr, event_ptr, pkt_ptr);
#endif /*GENUA*/
      if (!success)
	{
	  /* 
	   * There is a problem with the database. Set the global
	   * variable to inform the second thread to stop receiving 
           * data and close the connection
           */
	  DBFailed=1;

	  Printf("Error with database -> Call DumpData\n");

	  /* old code removed, we handle now DumpData via SIGINT */
	  kill(getpid(),SIGINT);
	}

#ifdef ALERT
      if ((UnixSocket) &&
	  (WTSdata->alertpkt.event.priority >= UnixSocketPriority))
	{
      /* first we remove all old data and then refill all values */
      bzero((void *)&alert,sizeof(AlertData));

      snprintf((char *) alert.alertmsg, ALERTMSG_LENGTH, 
               (char *) WTSdata->alertpkt.alertmsg);

      alert.sid=DBdata_ptr->sid;
      alert.cid=DBdata_ptr->cid-1; /* Was increased in Database() */
      alert.sensorIP=DBdata_ptr->sensorIP;
      alert.generator = WTSdata->alertpkt.event.sig_generator;
      alert.sig_id = WTSdata->alertpkt.event.sig_id;

      alert.srcip=WTSdata->alertpkt.iph.ip_src.s_addr;
      alert.dstip=WTSdata->alertpkt.iph.ip_dst.s_addr;

      alert.protocol=WTSdata->alertpkt.iph.ip_proto;
      if (alert.protocol==IPPROTO_TCP)
	{
	  alert.sport=WTSdata->alertpkt.tcph.th_sport;
	  alert.dport=WTSdata->alertpkt.tcph.th_dport;
          alert.Flag.TCP.x2=WTSdata->alertpkt.tcph.th_offx2;
          alert.Flag.TCP.flags=WTSdata->alertpkt.tcph.th_flags;
	}
      else if (alert.protocol==IPPROTO_UDP)
	{
	  alert.sport=WTSdata->alertpkt.udph.uh_sport;
	  alert.dport=WTSdata->alertpkt.udph.uh_dport;
	}
      else if (alert.protocol==IPPROTO_ICMP)
	{
	  alert.Flag.ICMP.ICMPType=WTSdata->alertpkt.icmp_code;
	  alert.Flag.ICMP.ICMPCode=WTSdata->alertpkt.icmp_type;
	}

      alert.priority=WTSdata->alertpkt.event.priority;
      alert.classification=WTSdata->alertpkt.event.classification;
      if (WTSdata->alertpkt.val & NOPACKET_STRUCT)
	alert.time=0;
      else
	alert.time=WTSdata->alertpkt.pkth.ts.tv_sec;
      /* Finally send it to the socket */
      send2socket((void *)&alert,sizeof(AlertData));
	}
#endif /*ALARM*/

      /* Remember the next pointer, after the free() call it would be lost */
      tmp=WTSdata->next;

      /* Release memory, the data is stored to the database, we can   */
      /* destroy the structure: first the payload and then the struct */
      if (WTSdata->pkt_ptr) free(WTSdata->pkt_ptr);
      free(WTSdata);

#ifdef DEBUG_WTS
      Printf("WTS> Data freed\n");
#endif

      /* Restore pointer to the next awaited data */
      WTSdata=tmp;
      WTScounter++;

#ifdef DEBUG_WTS
      Printf("WTS> WriteToSocket thread: %i\n",WTScounter);
#endif
  }

  /* Should never be reached... */
  pthread_exit(NULL);
}

/*
 * Function: ReceiveFromClient
 *
 * Purpose: Thread for reading snort alerts via tcp socket
 *         
 * Arguments: Pointer to the TCP socket 
 * 
 * Returns: Nothing, but only if client closes connections,
 *          otherwise it should never return...
 *
 * Author: DG
 */

void *
ReceiveFromClient(void *ptr)
{
  AlertpktDB alertpkt;
  static char buf[MAXBUFFSIZE];
  static char errbuf[STRLEN];
  int sock;
  unsigned int length,size;
  DGPacket *last;

#ifdef DEBUG_RFC
  char * text;
#endif
  
  sock= *(int *) ptr;

#ifdef DEBUG_RFC
  Printf("Start RFC\n");
#endif

  RFCthread=pthread_self();
#ifdef DROP
  /* store thread number for drop on SIGINT */
  /* Signal(SIGINT,SigDrop); */
  /* Signal(SIGTERM,SigDrop); */
#endif

  /* starting endless loop */
  while (1) 
    {  

#ifdef DEBUG_RFC
      Printf("RFC> Before recv\n");
#endif
      size=sizeof(AlertpktDB);
      bzero(&alertpkt,size);

      /* Check if the Database failed, if so close the connection */
      if (DBFailed) 
	{
	  clientexit=1;
	  close(sock);

#ifdef DEBUG_RFC
	  Printf("RFC> We closed connection, the database is gone...\n");
	  Printf("RFC> UNLOCKING mutex and sendig singal to exit\n");
#endif
	  pthread_cond_signal(&thr_cond);
	  pthread_mutex_unlock(&mutex);
	  
	  pthread_exit(NULL);
	}

      if ((length=DGrecv(sock,(void *) &alertpkt,size,MSG_WAITALL)) == -1)
	{
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Reading stream message: %s\n",errbuf);
	}
      if (length != size) 
	{
	  /* There seems to be something wrong with the client */
          /* Maybe the client closed the connection?           */
#ifdef DEBUG_RFC
	  Printf("RFC> WARNING: Read size (%i) is different"
		 " to structure size (%i)\n",length,size);
	  DG_strerror_r(errno,errbuf,STRLEN);	   
	  Printf("RFC> Ignoring entry: %s\n",errbuf);
#endif
	  if (length)
            {
	       Printf("WARNING: Read size (%i) is different"
	              " to structure size (%i)\n",length,size);
	       DG_strerror_r(errno,errbuf,STRLEN);	   
	       Printf("Ignoring entry: %s\n",errbuf);
	    }
	  break;  /* exit loop */
	}

#ifdef DEBUG_RFC
      Printf("RFC> Alarm        : %u\n",RFCcounter);
#ifndef GENUA
      Printf("RFC> Alert-Message: %s\n",alertpkt.alertmsg);
#endif
      Printf("RFC> Alert-Id     : %i\n",alertpkt.event.event_id);
      Printf("RFC> Priority     : %i\n",alertpkt.event.priority);
#endif

      if(!(alertpkt.val & NOPACKET_STRUCT)) 
	{
	  /* This is an alert with packet structure, not all have such */
	  /* a structure, e.g. portscans have only an alert message    */
	  size=alertpkt.pkth.caplen;
	  bzero(&buf,size);

	  length=DGrecv(sock,(void *) &buf,alertpkt.pkth.caplen,MSG_WAITALL);
	  if (length != size) 
	    {
	      /* There seems to be something wrong with the client */
	      /* Maybe the client closed the connection?           */
#ifdef DEBUG_RFC
	      Printf("RFC> WARNING: Read size (%i) is different"
		     " to packet size (%i)\n",length,size);
	      DG_strerror_r(errno,errbuf,STRLEN);	   
	      Printf("RFC> Ignoring entry: %s\n",errbuf);
#endif
	      if (length)
                {
	           Printf("WARNING: Read size (%i) is different"
		          " to packet size (%i)\n",length,size);
	           DG_strerror_r(errno,errbuf,STRLEN);	   
	           Printf("Ignoring entry: %s\n",errbuf);
		}
	      break; /* exit loop */
	    }

#ifdef DEBUG_RFC
	  text=Fasthex((u_char *)&buf,alertpkt.pkth.caplen);
	  Printf("RFC> Paketsize: %i, read: %i\n",alertpkt.pkth.caplen,length);
	  Printf("RFC> Content: %s\n",text);
#endif
	  /* allocate memory for the sniffed packet */
	  data->pkt_ptr = malloc(alertpkt.pkth.caplen);

	  if (alertpkt.data) 
	    {
	      /* Calculate Pointer to real payload of the packet */
	      alertpkt.data += (int) data->pkt_ptr;
	    }
#ifdef DEBUG_RFC
	  Printf("RFC> Payload at %i (%p)\n",
		 alertpkt.data,data->pkt_ptr);
#endif
	  /* Store the captured packet in pointer shared with */
          /* the other thread, this is simply to keep the buf */
          /* buffer free for the next packet. The idea is to  */
          /* simply keep the buf and drop (or save to file)   */
          /* the data in the case the above malloc fails...   */
	  bcopy(&buf, data->pkt_ptr, alertpkt.pkth.caplen);
	}
      
      /* Now we have to set the mutex lock. This is to ensure that */
      /* we do not use an incomplete structure in the other thread */

      pthread_mutex_lock(&mutex);

      /* Allocate memory for the next packet */
      data->next = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
      last = data; /* Pointer to actual data becomes last in next data...*/
      data->alertpkt=alertpkt;
      data=data->next;
      data->prev=last; /* last pointer in new packet shows to old data */
      RFCcounter++;
#ifdef DROP
      if (DropSocket && 
	  ((RFCcounter+RSTcounter-DropCounter-WTScounter) > HighWater)) 
	{
#ifdef DEBUG_DROPPACKETS
	  Printf("Diff: %i (%i)\n",(RFCcounter+RSTcounter-DropCounter-WTScounter),
		 HighWater);
#endif
	  DropPackets(last);
	}
#endif

#ifdef DEBUG_RFC
      Printf("RFC> Data      at %p\n",data);
      Printf("RFC> Data next at %p\n",data->next);
      Printf("RFC> Send signal!\n");
#endif

      /* Now we can inform the other thread that there is a */
      /* complete new packet available which can be send to */
      /* the database */
      pthread_cond_signal(&thr_cond);

#ifdef DEBUG_RFC
      Printf("RFC> Sent signal!\n");
#endif

      pthread_mutex_unlock(&mutex);
      
    }
  /* If this point is reached then there is something wrong with  */
  /* the client: Close the connection and set clientexit so that  */
  /* the other thread is informed about the situation and should  */
  /* exit as soon as all remaining data is stored to the database */
  clientexit=1;
  close(sock);

  Printf("Client closed connection, we too...\n");
#ifdef DEBUG_RFC
  Printf("RFC> UNLOCKING mutex and sendig singal to exit\n");
#endif

  pthread_cond_signal(&thr_cond);
  pthread_mutex_unlock(&mutex);

  pthread_exit(NULL);
}


/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  Printf("\nUsage: %s [-bhflnpTuv] [-A delay ] [-M priority] [-m mode] "
	 "[-p port] [-P PIDFile] [-s name] [-S server] [-w dir] "
         "[-W swapdir]\n",argv);
  Printf("\n -A AlarmDelay : interval in seconds to display statistics");
  Printf("\n -b            : start in daemon mode, go in background");
  Printf("\n -c config     : configuartion file");
  Printf("\n -d            : dump config to stdout");
  Printf("\n -f            : store full payload including all headers");
#ifdef DROP
  Printf("\n -D dropsocket : socket name for dropping alerts");
#endif
  Printf("\n -h            : this information");
#ifdef DROP
  Printf("\n -H highwater  : high water mark for dropping alerts");
#endif
  Printf("\n -l            : print statistics via syslogd");
#ifdef DROP
  Printf("\n -L lowwater   : low water mark for dropping alerts");
#endif
  Printf("\n -m mode       : umask value for daemon mode\n");
#ifdef ALERT
  Printf("\n -M priority   : minimum priority to send alerts to unix socket");
#endif
  Printf("\n -n            : do not try to resolve FQN of sensor");
  Printf("\n -p port       : TCP port database server listens on");
  Printf("\n -P PIDfile    : filename to store pid in");
  Printf("\n -s socketname : name of unix socket to be created");
  Printf("\n -S server     : name of database server");
  Printf("\n -T            : trust that signatures are already in database");
#ifdef ALERT
  Printf("\n -u            : opposite of -U, disables use of alert socket");
  Printf("\n -U alertsocket: socket name for alerts");
#endif
  Printf("\n -v            : version information");
  Printf("\n -w daemondir  : directory to change to in daemon mode");
  Printf("\n -W swapdir   : directory to temporarily store alerts if database is gone.");
  Printf("\n\n");
  exit(0);
}

/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */
void 
CmdLine(int argc, char* argv[]) {
  int opt, size, i;
  char options[]="bfm:nhvw:c:s:S:p:A:tTU:dlM:P:uD:L:H:W:";

  while((opt = getopt(argc, argv, options)) != -1) 
    {
      switch(opt)
	{
	case 'v':
	  Printf("\nThreaded socket server to create TCP socket, read\n");
	  Printf("data from remote clients and transfer them via unix\n");
          Printf("socket to a database.\n");
	  Printf("\nWritten by Dirk Geschke\n\n");
	  Printf("%s\n\n",rcsid);
	  Printf("Version %s (snort-%s)\n\n",FLoP_VERSION,VERSION);
	  exit(0);
	  break;
      
	case 'h':
	  Usage(argv[0]);
	  break;
	  
	case 'm':
	  mask=(mode_t) strtol(optarg,NULL,0);
	  maskSet=1;
	  break;

	case 'c':
	 size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	 strncpy(ConfigFile,optarg, size);
	 ConfigFile[size]='\0';
#ifdef DEBUG_CMDLINE
	 Printf("ConfigFile: %s\n",ConfigFile); 
#endif
	  break;

	case 'f':
	  FullPayload = 1;
#ifdef DEBUG_CMDLINE
	 Printf("Store full payload: activated\n"); 
#endif
	  break;

	case 's':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(SocketName,optarg, size);
	  SocketName[size]='\0';
	  SocketNameSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("Socket: %s\n",SocketName);
#endif
	  break;  
	  
	case 'S':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(ServerName,optarg, size);
	  ServerName[size]='\0';
	  ServerNameSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("Server: %s\n",ServerName);
#endif
	  break;  

#ifdef ALERT
	case 'U':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(UnixSocketName,optarg, size);
	  UnixSocketName[size]='\0';
	  if (strlen(UnixSocketName)) UnixSocket=1;
#ifdef DEBUG_CMDLINE
	  Printf("UnixSocket: %s\n",UnixSocketName);
#endif
	  break;
#endif  

#ifdef DROP
	case 'D':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  if (size > UNIX_PATH_MAX)
	    {
	      Printf("Length of SocketName too large!\n");
	      Printf("Maximum length: %u\n",UNIX_PATH_MAX);
	      Printf("Actual length : %u\n",size);
	      Printf("Exiting.\n");
	      exit(1);
	    }
	  strncpy(DropSocketName,optarg, size);
	  DropSocketName[size]='\0';
	  if (strlen(DropSocketName)) DropSocket=1;
#ifdef DEBUG_CMDLINE
	  Printf("DropSocket: %s\n",DropSocketName);
#endif
	  break;
	  
	case 'L':
	  LowWater=atoi(optarg);
	  LowWaterSet=1;
	  break;

	case 'H':
	  HighWater=atoi(optarg);
	  HighWaterSet=1;
	  break;
#endif /* DROP */
	  
	case 'p':
	  ServerPort=atoi(optarg);
	  ServerPortSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("ServerPort: %i\n",ServerPort);
#endif
	  break;  
	  
	case 'A':
	  i=atoi(optarg);
          if (i > 0 ) AlarmDelay=i;
	  else AlarmDelay=0;
	  AlarmDelaySet=1;
#ifdef DEBUG_CMDLINE
	  Printf("AlarmDelay: %i\n",AlarmDelay);
#endif
	  break;

#ifdef ALERT
	case 'u':
	  UnixSocket=0;
	  UnixSocketSet=1;
	  break;

	case 'P':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(PIDFile,optarg, size);
	  PIDFile[size]='\0';
	  PIDFileSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("PIDFile: %s\n",PIDFile);
#endif
	  break;  

	case 'M':
	  UnixSocketPriority=atoi(optarg);
	  UnixSocketPrioritySet=1;
#ifdef DEBUG_CMDLINE
	  Printf("UnixSocket priority: %i\n",UnixSocketPriority);
#endif
	  break;
#endif

        case 'T':
          DBtrust=1;
          DBtrustSet=1;
#ifdef DEBUG_CMDLINE
          Printf("DBtrust is set.\n");
#endif
          break;

        case 't':
          DBtrans=1;
          DBtrans=1;
#ifdef DEBUG_CMDLINE
          Printf("DBtrust is set.\n");
#endif
          break;

        case 'n':
          FQNSensor=0;
          FQNSensorSet=1;
#ifdef DEBUG_CMDLINE
          Printf("FQN for sensor disabled.\n");
#endif
          break;

        case 'd':
          DumpConf=1;
#ifdef DEBUG_CMDLINE
          Printf("Dump configuration.\n");
#endif
          break;

        case 'l':
          LogSyslog=1;
#ifdef DEBUG_CMDLINE
          Printf("Dump statistics to syslogd instead stdout.\n");
#endif
          break;

        case 'b':
          Daemon=1;
	  DaemonSet=1;
	  LogSyslog=1;
#ifdef DEBUG_CMDLINE
          Printf("Starting in daemon mode is set.\n");
#endif
          break;

	case 'w':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(DaemonDir,optarg, size);
	  DaemonDir[size]='\0';
	  DaemonDirSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("Working directory for daemon mode: %s\n",DaemonDir);
#endif
	  break;  

	case 'W':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(SwapDir,optarg, size);
	  SwapDir[size]='\0';
	  SwapDirSet=1;
#ifdef DEBUG_CMDLINE
	  Printf("Working directory for swapping: %s\n",SwapDir);
#endif
	  break;  

	default:
	  Printf("Unknown flag: %c\n",opt);
	  Usage(argv[0]);
	}
    }
}

/* 
 * Function: OpenServer
 *
 * Purpose: Opens tcp server in listening mode
 * 
 * Arguments: servername for binding on IP-Number
 *              (it would be wise to use an IP-Number instead 
 *               of an full qualified name in case of more than
 *               one IP address is associated with the FQN!)
 *              "0.0.0.0" means any address, e.g. bind on all
 *              interfaces.
 *            port, on which port to bind the service.
 *
 * Returns: socket to wait for incoming client requests
 *
 * Author: DG
 *
 */
int 
OpenServer(char *servername, int port)
{
#ifdef HAVE_BIND_HEADER
  int errnum;
#endif
  int sock,i;
  struct hostent *server_ptr;
  struct sockaddr_in srvin;
  char errbuf[STRLEN];
    
  bzero((void *)&srvin, sizeof(srvin));
  
  srvin.sin_port = htons((u_short) port);
  srvin.sin_family= AF_INET;

#ifdef DEBUG_OS
  Printf("OS> Port: %i (%i)\n",port, srvin.sin_port);
#endif

  /* get a socket */
  sock=socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0 ) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);	   
      Printf("Error creating socket: %s\n",errbuf);
      exit(1);
    }

  /* Set keepalive option for an existing connection not to    */
  /* close if actually no data is received. This is not really */
  /* necessary, otherwise we would restart a new connection    */
  i=1;
  if (setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, (void *) &i,sizeof(i)) ==-1)
    {
      DG_strerror_r(errno,errbuf,STRLEN);	   
      Printf("Setting SO_KEEPALIVE failed: %s\n",errbuf);
    }

  /* Set reuseaddr option for a quick reuse of the used addresses */
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *) &i,sizeof(i)) ==-1)
    {
      DG_strerror_r(errno,errbuf,STRLEN);	   
      Printf("Setting SO_REUSEADDR failed: %s\n",errbuf);
    }
	     
  /* get address of server */
#ifdef HAVE_BIND_HEADER
  server_ptr=getipnodebyname(ServerName, AF_INET, AI_ALL, &errnum);
#else
  server_ptr=gethostbyname(ServerName);
#endif
  if (server_ptr == NULL) 
    {
      Printf("%s: unknown hostname\n",ServerName);
      exit(1);
    }

  memcpy((void *)&srvin.sin_addr, server_ptr->h_addr_list[0], 
	 server_ptr->h_length);

  /* Now bind the socket to an interface */
  if (bind(sock, (struct sockaddr *) &srvin, sizeof(srvin)) < 0) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error bind: %s\n",errbuf);
      exit(1);
    }

#ifdef DEBUG_OS
  Printf("OS> bound socket.\n");
#endif

  /* Now start to listen on the bounded tcp socket */
  if (listen(sock,SOMAXCONN)<0) 
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error listen: %s\n",errbuf);
      exit(1);
    }

  /* free the memory needed for name resolution */
#ifdef HAVE_BIND_HEADER
  freehostent(server_ptr); 
#endif

#ifdef DEBUG_OS
  Printf("OS> Listening on port %i\n",port);
#endif

  return sock;
}

/* Function: ReadConfig()
 *
 * Purpose: read configuration files to set at least the parameter
 *          for database access (avoid password on command line)
 *          spaces/tabs at the beginning are removed, spaces at
 *          the end too. But maybe we should use a better parser...
 *
 * Author: DG
 * 
 */

void 
ReadConfig()
{
  FILE *fd = NULL;
  char *token =NULL;
  char *arg=NULL;
  char *val=NULL;
  char line[STRLEN];
  int linenumber=0;
  int i,len;

  if ( (fd=fopen(ConfigFile,"r")) == NULL ) 
    {
      DG_strerror_r(errno,line,STRLEN);	   
      Printf("Can not open config file \"%s\": %s\n", ConfigFile, line);
      exit(1);
    }
  while((fgets(line, STRLEN, fd))!=NULL ) 
    { 
      linenumber++;

      /* remove end of line */
      if ( (token = strchr (line,'\n')) ) *token='\0';

      /* skip if line starts with a comment: if no comment sign */
      /* is found we are sure that we can derefence the pointer */
      /* before any comment sign!*/

      if ( line[0]=='#' ) continue;

      /* remove any comments at end of line and if a comment */
      /* sign is escaped: remove backslash from string!      */
      token = line;
      while ( (token = strchr(token,'#')) && (*(token-1)=='\\' ) )
      {
         val=token;
         arg=token-1;
	 while (*arg) {
		 *arg=*token;
		 arg++;
		 token++;
	 }
	 token=val;
      }

      /* Maybe token is pointing now to a real comment sign */
      if (token && *token=='#') *token='\0';

      /* next if line empty (or only a comment) */
      if (line[0]=='\0') continue;

      /* Up to ":" or "=" we get the argument */
      if ( (strstr(line,":") == NULL) && (strstr(line,"=") == NULL) )
        {
           Printf("\nIgnoring unknown entry in config file: \n");
           Printf("\n> %s (missing : or = ?)\n\n",line);
           continue;
        }

      arg = strtok(line,":=");
      /* second argument is the value, remove tabs */
      val = strtok(NULL,"\0");
      /* remove leading spaces or tabs */
      while ( arg[0]==' ' || arg[0]=='\t') arg++;
      while ( val[0]==' ' || val[0]=='\t') val++;
      /* remove spaces at end of value */
      while ( (strlen(arg) > 1) && (arg[strlen(arg)-1] == ' ') )
		    arg[strlen(arg)-1]='\0';
      while ( (strlen(val) > 1) && (val[strlen(val)-1] == ' ') )
		    val[strlen(val)-1]='\0';
#ifdef DEBUG_READCONFIG
      Printf("Line: %i Argument: <%s> Value: <%s>\n",linenumber,arg,val);
#endif
      ArgCheck(val);

      len=strlen(val);
      if(len > STRLEN-2) 
	{
	  Printf("Value in config file too long: %s\n",val);
	  exit(1);
	}

      if ( !strncasecmp(arg,"dbuser",6) )
        {
           strncpy(DBuser,val,len);
	   DBuser[len]='\0';
	   continue;
        } 

      if ( !strncasecmp(arg,"dbpassword",10) )
	{
	      strncpy(DBpassword,val,len);
	      DBpassword[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"dbname",6) )
	{
	      strncpy(DBname,val,strlen(val));
	      DBname[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"dbencoding",10) )
	{
	      strncpy(DBencoding,val,len);
	      DBencoding[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"fullpayload",11) )
	{
	      FullPayload=1;
	      continue;
	}

      if ( !strncasecmp(arg,"PIDFile",7) )
        {
	  if(!PIDFileSet)
	    {
           strncpy(PIDFile,val,len);
	   PIDFile[len]='\0';
	    }
	   continue;
        } 

      if ( !strncasecmp(arg,"serverport",10) )
	{
	      if (!ServerPortSet) ServerPort=atoi(val);
	      continue;
	}

      if ( !strncasecmp(arg,"servername",10) )
	{
	  if (!ServerNameSet) 
	    {
	      strncpy(ServerName,val,len);
	      ServerName[len]='\0';
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"socketname",10) )
	{
	  if (!SocketNameSet) 
	    {
	      if (len > UNIX_PATH_MAX)
		{
		  Printf("Length of SocketName too large!\n");
		  Printf("Maximum length: %u\n",UNIX_PATH_MAX);
		  Printf("Actual length : %u\n",len);
		  Printf("Exiting.\n");
		  exit(1);
		}
	      strncpy(SocketName,val,len);
	      SocketName[len]='\0';
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"alertsocket",11) )
	{
#ifdef ALERT
	  if ((!UnixSocket) && (!UnixSocketSet)) 
	    {
	      if (len > UNIX_PATH_MAX)
		{
		  Printf("Length of SocketName too large!\n");
		  Printf("Maximum length: %u\n",UNIX_PATH_MAX);
		  Printf("Actual length : %u\n",len);
		  Printf("Exiting.\n");
		  exit(1);
		}
	      UnixSocket=1;
	      strncpy(UnixSocketName,val,len);
	      UnixSocketName[len]='\0';
	    }
#else
          Printf("Option AlertSocket not enabled!\n");
	  Printf("Recompile with -DALERT\n");
#endif
	  continue;
	}

      if ( !strncasecmp(arg,"UnixSocketPriority",18) )
	{

#ifdef ALERT
	  if (!UnixSocketPrioritySet) 
	    {
	      UnixSocketPriority=atoi(val);
	    }
#else
          Printf("Option UnixSocketPriority not enabled!\n");
	  Printf("Recompile with -DALERT\n");
#endif
	  continue;
	}

      if ( !strncasecmp(arg,"dbtrust",7) )
        {
          if (!DBtrustSet) DBtrust=atoi(val);
          continue;
        }

      if ( !strncasecmp(arg,"dbtype",7) )
	{
#ifdef MYSQL_DB
	  if (!strncasecmp(val,"MYSQL",5))
	    { 
#ifdef DEBUG_DB
	      Printf("Database type: MySQL\n");
#endif
	      DBtype=DB_MYSQL;
	    }
#endif
#ifdef POSTGRES
	  if (!strncasecmp(val,"Postgres",8))
	    { 
#ifdef DEBUG_DB
	      Printf("Database type: Postgres\n");
#endif
	      DBtype=DB_POSTGRES ;
	    }
#endif
	  continue;
	}

      if ( !strncasecmp(arg,"dbtrans",7) )
	{
	  if (!DBtransSet) DBtrans=atoi(val);
	  continue;
	}

      if ( !strncasecmp(arg,"alarmdelay",10) )
	{
	  if (!AlarmDelaySet) 
	    {
	      i=atoi(val);
	      if ( i > 0) AlarmDelay=i;
	      else AlarmDelay=0;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"syslog",6) )
	{
	  if (!LogSyslog) LogSyslog=atoi(val);
	  continue;
	}

      if ( !strncasecmp(arg,"DaemonMode",10) )
	{
	  if (DaemonSet==0)
	    {
	      Daemon=atoi(val);
	      /* if daemon mode we have to use syslog */
	      if (Daemon) LogSyslog=1;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"fqnsensor",9) )
	{
	  if (!FQNSensorSet)
	    {
	      i=atoi(val);
	      if ( i > 0) FQNSensor=i;
	      else FQNSensor=0;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"umask",5) )
	{
	  if (!maskSet)
	    {
	      i=strtol(val,NULL,0);
	      if ( i > 0) mask=i;
	      else mask=0;
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"DaemonDir",5) )
	{
	  if (!DaemonDirSet)
	    {
	      strncpy(DaemonDir,val,len);
	      DaemonDir[len]='\0';
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"SwapDir",7) )
	{
	  if (!SwapDirSet)
	    {
	      strncpy(SwapDir,val,len);
	      SwapDir[len]='\0';
	    }
	  continue;
	}

      if ( !strncasecmp(arg,"dropsocket",10) )
	{
#ifdef DROP
	  if ((!DropSocket) && (!DropSocketSet)) 
	    {
	      if (len > UNIX_PATH_MAX)
		{
		  Printf("Length of SocketName too large!\n");
		  Printf("Maximum length: %u\n",UNIX_PATH_MAX);
		  Printf("Actual length : %u\n",len);
		  Printf("Exiting.\n");
		  exit(1);
		}
	      DropSocket=1;
	      strncpy(DropSocketName,val,len);
	      DropSocketName[len]='\0';
	    }
#else
          Printf("Option DropSocket not enabled!\n");
	  Printf("Recompile with -DDROP\n");
#endif
	  continue;
	}

      if ( !strncasecmp(arg,"LowWater",8) )
	{

#ifdef DROP
	  if (!LowWaterSet) 
	    {
	      LowWater=atoi(val);
	    }
#else
          Printf("Option LowWater not enabled!\n");
	  Printf("Recompile with -DDROP\n");
#endif
	  continue;
	}

      if ( !strncasecmp(arg,"HighWater",9) )
	{

#ifdef DROP
	  if (!HighWaterSet) 
	    {
	      HighWater=atoi(val);
	    }
#else
          Printf("Option HighWater not enabled!\n");
	  Printf("Recompile with -DDROP\n");
#endif
	  continue;
	}

      Printf("\nWARNING: unknown option in line %i: %s\n",linenumber,arg);
      
    }
  fclose(fd);
}


/* Function: DumpConfig()
 *
 * Purpose: dump all relevant settings on runtime, password for
 *          database access is not printed.
 *
 * Author: DG
 *
 */
void 
DumpConfig()
{
    Printf("\n Configuration parameter:\n");
    Printf(" ========================\n");
    Printf(" ConfigFile : <%s>\n",ConfigFile);
    Printf(" PIDFile    : <%s>\n",PIDFile);
    Printf(" SocketName : <%s>\n",SocketName);
    Printf(" ServerName : <%s>\n",ServerName);
    Printf(" ServerPort : <%i>\n",ServerPort);
    Printf(" AlarmDelay : <%i>\n",AlarmDelay);
    if (LogSyslog)
      {
         Printf(" Logging    : syslog\n");
      }
    else
      {
	 Printf(" Logging    : stdout\n"); 
      }
#ifdef ALERT
    if (UnixSocket)
      {
	Printf(" AlarmSocket: <%s>\n",UnixSocketName);
	Printf(" US-priority: <%i>\n",UnixSocketPriority);
      }
    else
      Printf(" AlarmSocket: disabled\n");
#endif
    Printf(" ConfigFile : <%s>\n",ConfigFile);
    Printf(" DBname     : <%s>\n",DBname);
    Printf(" DBuser     : <%s>\n",DBuser);
    if (DBpassword) 
      {
        Printf(" DBpassword : is set (length: %i Bytes)\n",strlen(DBpassword));
      }
    else 
      {
        Printf(" DBpassword : is not set (empty password!)\n");
      }
    Printf(" DBtrans    : <%i>\n",DBtrans);
    Printf(" DBtrust    : <%i>\n",DBtrust);
    Printf(" DBtype     : <%i>\n",DBtype);
    Printf(" DBencoding : <%s>\n\n",DBencoding);
    if (FullPayload)
      Printf(" FullPayload: enabled\n");
#ifdef DROP
    if(DropSocket)
      {
	Printf(" Dropping is enabled.\n");
	Printf(" DropSocket : <%s>\n",DropSocketName);
	Printf(" LowWater   : <%u>\n",LowWater);
	Printf(" HighWater  : <%u>\n",HighWater);
      }
    else
      Printf(" Dropping is disabled.\n");
#endif
}

/* Function: SigChld
 *
 * Purpose: capture SIGCHILD to avoid zombies in the process table
 *          in addition we clean up the child process list. This 
 *          list is used to gather the statistics from all childs.
 *
 * Author: DG
 */

void 
SigChld()
{
  int i,status=0;
  pid_t pid;

  /* get the pid to avoid zombies, we ignore the status yet... */
  pid = wait(&status);

  /* Remove pid from list (if in list...) */
  for (i=0; i<MAXCHILD; i++)
    {
      if (Child[i].pid==pid) 
	{
	  Child[i].pid=0;
	  Child[i].ip=0;
	  break;
	}
    }
#ifdef DEBUG_CHLD
  Printf("\n Child exit: %i\n",pid);
#endif  

  return;
}

/* Function: SigUsr1()
 *
 * Purpose: The main process sends SigUsr2 to all child processes
 *          to gather all statistics (otherwise you have to send a
 *          signal to each child process which are difficult to 
 *          identify if you use threads (at least under linux each
 *          thread gets an two processes: one in user space and one
 *          in kernel space
 *
 *          child processes should ignore 1
 *
 * Author: DG
 *
 */
void 
SigUsr1()
{
  int i;
   
  if(!LogSyslog)
    {
       printf("\n--- Statistics from clients --- %s\n",Time());
      /* Flush buffer to get first the line and then the statistics... */
       fflush(stdout);
    }
  for (i=0; i<MAXCHILD; i++)
    {
      if (Child[i].pid) 
	{
	  /* we have a child process found and send USR2 to it */
	  kill(Child[i].pid,SIGUSR2);
	}
    }
}

/* Function: SigUsr2()
 *
 * Purpose: The main process sends SigUsr2 to all child processes,
 *          the child processes write then the statistics to stdout
 *
 * Author: DG
 */
void 
SigUsr2()
{
  pid_t pid;
  static u_int RFC_last=0;
  static u_int WTS_last=0;
  static u_int Drop_last=0;

#ifdef ONE_STAT_THREAD
  /* only one thread should display statistics */
  if ( !pthread_equal(RFCthread,pthread_self()) )  return;
#endif

  pid=getpid();

  /*  Printf("\nChild: %i\n",pid); */
  Printf("[%u]Entries received: %i (%i)\n", (u_int) pid, RFCcounter, 
	 RFCcounter-RFC_last);
  Printf("[%u]Entries send    : %i (%i)\n", (u_int) pid, WTScounter, 
	 WTScounter-WTS_last);
  Printf("[%u]Entries dropped : %i (%i)\n", (u_int) pid, DropCounter, 
	 DropCounter-Drop_last);
  if (RSTcounter)
    Printf("[%u]Entries restored: %i\n", (u_int) pid, RSTcounter);

  RFC_last=RFCcounter;
  WTS_last=WTScounter;
  Drop_last=DropCounter;
}

/*
 * Function: SigAlarm()
 * 
 * Purpose: Catch ALARM-Signal to print the statistics to stdout
 *          and do a reset of the alarm handler to get statistics
 *          on a periodically interval
 *
 * Author: DG
 *
 */
void 
SigAlarm()
{
  SigUsr1();
  alarm(AlarmDelay);
}

/*
 *  Function: SigInt()
 *  
 *  Purpose: Catch INT-Signal to print the statistics to stdout
 *           and do a clean exit
 * 
 *  Author: DG
 *
 */
void 
SigInt()
{
  /* Disable alarm function */
  alarm(0);
  SigUsr1();
  unlink(PIDFile);
  _exit(0);
}

/*
 *  Function: SigInt2()
 *  
 *  Purpose: Catch INT-Signal to exit the child threads
 * 
 *  Author: DG
 *
 */
void 
SigInt2()
{
  int ret;
#ifdef ONE_SIG_THREAD
  static int called=0;
#endif

#ifdef DEBUG
  Printf("SigInt2 called, thread: %i (%i)\n",pthread_self(),WTSthread);
#endif

#ifdef ONE_SIG_THREAD
  if (called) { pthread_exit(0); }
  called++;
#else 
  if ( pthread_equal(WTSthread,pthread_self()) ) 
#endif
    {
#if defined DEBUG 
#if defined ONE_SIG_THREAD
      Printf("Closing single thread\n");
#else 
      Printf("WTSthread is running\n");
#endif
#endif
      ret=DumpData(SwapFile,WTSdata);
      if (ret < 0)
	{
	  /* DumpData failed, so we call DROP*/
#ifdef DROP
	  LowWater=0;
	  DropPackets(WTSdata);
#endif
	}
      else
	Printf("Saved alerts in DumpData: %i\n",ret);
      Disconnect(DBdata);
    }
  
#ifdef ONE_SIG_THREAD
  _exit(0);
#else
  pthread_exit(0);
#endif
}

/*
 *  Function: SigDrop()
 *  
 *  Purpose: Catch INT-Signal dump remaining alerts to drop socket
 *           and do a clean exit
 * 
 *  Author: DG
 *
 */
#ifdef DROP
void 
SigDrop()
{
  if ( pthread_equal(RFCthread,pthread_self()) ) 
    {
      LowWater=0;
      DropPackets(data);
    }
  _exit(0);
}
#endif


/*
 *  Function: SigHup()
 *  
 *  Purpose: Catch HUP-Signal, stop all forked processes
 *           and do a restart
 * 
 *  Author: DG
 *
 */
void 
SigHup()
{
  sigset_t sigset,osigset;
  int i;

  /* Disable alarm function */
  alarm(0);
  Printf("SIGHUP received!\n");

  if ( pthread_equal(MainThread,pthread_self()) )
    {
      SigUsr1();
      /* okay, we ignore SIGCHLD, otherwise we were taken to another */
      /* routine on each kill. We wait for the pid to avoid zombies  */
      Signal(SIGCHLD,SIG_IGN);
      /*      Disconnect(DBdata); -> child process must close connection */
      unlink(PIDFile);
      close(ServerSocket);
      Printf("Terminating child processes:\n");
      for (i=0; i<MAXCHILD; i++)
	{
	  if (Child[i].pid) 
	    {
	      Printf("Kill child: %u\n",Child[i].pid);
	      /* we have a child process found and send TERM to it */
	      kill(Child[i].pid,SIGTERM);
	      /* wait for process to finish to avoid zombies */
	      waitpid(Child[i].pid, NULL,0);
	      Child[i].pid=0;
	    }
	}

      Printf("Restarting %s\n",prog);

      /* Reenable SIGHUP, we were called by signal handler for SIGHUP */
      /* so normally this signal is blocked during this routine and   */
      /* this behaviour is inherited by the execv-process             */
      sigaddset(&sigset,SIGHUP);
      sigprocmask(SIG_UNBLOCK,&sigset,&osigset);
      
      /* So finally restart the main program */
      execv(prog,args);
    }
  /* never reached except if execv fails */
  _exit(1); 
}


/* 
 *  Function: ArgCheck
 *
 *  Purpose: Check for occurence of wrapper around the arguments
 *           in the config file. We remove ' and " if they are at
 *           the beginning and the end.
 *
 *  Arguments: the extracted value as string
 *
 *  Author: DG
 *
 */

void 
ArgCheck(char *arg)
{
  char Value[STRLEN];
  int len;
  
  len=strlen(arg);

  if (len < 2) return;
#ifdef DEBUG_READCONFIG
  Printf("Arg: %s, Len: %i\n",arg,len);  
#endif

  if ( ((index(arg,'"')==arg) && (rindex(arg,'"')==arg+len-1)) ||
       ((index(arg,'\'')==arg) && (rindex(arg,'\'')==arg+len-1)) )
   {
       strncpy(Value,arg+1,len-2);
       /* dont't forget to end the string... */
       Value[len-2]='\0';
#ifdef DEBUG_READCONFIG
       Printf("Value: %s\n",Value);
#endif
       strcpy(arg,Value);
   }
  return;
}


/*
 * Function: Time()
 * 
 * Purpose: returns actual time in readable format without
 *          ending \n
 *
 * Author: DG
 *
 */

char *
Time()
{
  static char message[STRLEN];
  time_t timep;

  timep=time(NULL);
  strcpy(message,ctime(&timep));
  message[strlen(message)-1]='\0'; 
  return message;
}


/*
 * Function: CheckIP(u_int32_t)
 * 
 * Purpose: Checks if a sensor is already connected and 
 *          do a cleanup if there is no running process
 *          according to the pid.
 *          Note: Even if a sensor is disconnected it 
 *                is still possible that the process 
 *                is running due to remaining data not
 *                stored in the database yet.
 *
 * Returns: 0: sensor is not connected
 *          1: there is a sensor already connected or 
 *             remaining data not yet stored int the
 *             database
 *
 * Author: DG
 *
 */

int 
CheckIP(u_int32_t ip)
{
  int i;
  int result;

  result=0;
  for (i=0; i<MAXCHILD; i++)
    {
      if (Child[i].ip==ip)
	{
	  /* Sensor is already connected */
	  result=1;
	  
	  /* Test if the child process is running */
	  if ( kill(Child[i].pid,0))
	    {
	      /* Child process is not running, clear entry */
	      Child[i].pid=0;
	      Child[i].ip=0;
	      result=0;
	    }
	}
    }
  return result;
}


/*
 * Function: OpenAlertSocket(filename);
 * 
 * Purpose: Opens an existing unix domain socket (for writing)
 *          If the socket does not exist a warning is printed
 *          but the program does not stop.
 *
 * Returns: file descriptor of socket
 *
 * Author: DG
 *
 */
#ifdef ALERT
void 
OpenAlertSocket(char *filename)
{
  /*  struct sockaddr_un unix_addr;*/
  char errbuf[STRLEN];

#ifdef DEBUG_OAS
  Printf("OpenAlertSocket> SocketName: %s\n",filename);
#endif

  if (filename==NULL) 
    {
      Printf("Filename missing in OpenAlertSocket!\n");
      exit(1);
    }

  if(access(filename, W_OK))
    {
      Printf("WARNING: %s file doesn't exist or isn't writable!\n", filename);
    }

  bzero((char *) &alert_addr, sizeof(alert_addr));
  bcopy((const void *)filename, (void *)alert_addr.sun_path, strlen(filename));

  alert_addr.sun_family = AF_UNIX;
  
  if((UnixSocket = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("socket call failed: %s\n",errbuf);   
      exit(1);
    }
#ifdef DEBUG_OAS
   Printf("OpenAlertSocket> UnixSocket: %i\n",UnixSocket);
#endif
}
#endif /*ALERT*/

/*
 * Function: send2socket(data, size)
 * 
 * Purpose: sends data of size bytes to (unix domain) socket
 *
 * Returns: -1 on error
 *
 * Author: DG
 *
 */
#ifdef ALERT 
int 
send2socket(void * data, int size)
{
  char errbuf[STRLEN];

  if(sendto(UnixSocket,(const void *) data, size,
	    0,(struct sockaddr *)&alert_addr,sizeof(alert_addr))==-1)
    /*  if(send(UnixSocket,(const void *) data, size, 0)==-1)*/
    {
      DG_strerror_r(errno,errbuf,STRLEN);	   
      Printf("Error writing to unix socket <%s>: %s\n",UnixSocketName,errbuf); 
      return -1;
    }
  return 0;
}
#endif


/*
 * Function: drop2socket(data, size)
 * 
 * Purpose: sends data of size bytes to (unix domain) socket
 *
 * Returns: -1 on error
 *
 * Author: DG
 *
 */
#ifdef DROP
int 
drop2socket(void * data, int size)
{
  char errbuf[STRLEN];

  if(sendto(DropSocket,(const void *) data, size,
	    0,(struct sockaddr *)&drop_addr,sizeof(drop_addr))==-1)
    /*  if(send(UnixSocket,(const void *) data, size, 0)==-1)*/
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error writing to drop socket <%s>: %s\n", DropSocketName,errbuf);
      return -1;
    }
  return 0;
}
#endif


#ifdef DROP
/*
 * Function: dropAlert(data)
 * 
 * Purpose: writes alerts to unix socket, part of drop feature
 *
 * Author: DG
 *
 */

void 
DropAlert(DGPacket *data)
{
  AlertData alert;

  /* first we remove all old data and then refill all values */
  bzero((void *)&alert,sizeof(AlertData));

  snprintf((char *)alert.alertmsg, ALERTMSG_LENGTH, 
           (char *)data->alertpkt.alertmsg);
  
  alert.sid=0; /* not in database! */
  alert.cid=0;
  alert.sensorIP=DBdata->sensorIP;
  alert.generator = data->alertpkt.event.sig_generator;
  alert.sig_id = data->alertpkt.event.sig_id;

  alert.srcip=data->alertpkt.iph.ip_src.s_addr;
  alert.dstip=data->alertpkt.iph.ip_dst.s_addr;

  alert.protocol=data->alertpkt.iph.ip_proto;
  if (alert.protocol==IPPROTO_TCP)
    {
      alert.sport=data->alertpkt.tcph.th_sport;
      alert.dport=data->alertpkt.tcph.th_dport;
      alert.Flag.TCP.x2=data->alertpkt.tcph.th_offx2;
      alert.Flag.TCP.flags=data->alertpkt.tcph.th_flags;
    }
  else if (alert.protocol==IPPROTO_UDP)
    {
      alert.sport=data->alertpkt.udph.uh_sport;
      alert.dport=data->alertpkt.udph.uh_dport;
    }
  else if (alert.protocol==IPPROTO_ICMP)
    {
      alert.Flag.ICMP.ICMPType=data->alertpkt.icmp_code;
      alert.Flag.ICMP.ICMPCode=data->alertpkt.icmp_type;
    }
  
  alert.priority=data->alertpkt.event.priority;
  alert.classification=data->alertpkt.event.classification;
  if (data->alertpkt.val & NOPACKET_STRUCT)
    alert.time=0;
  else
    alert.time=data->alertpkt.pkth.ts.tv_sec;
  /* Finally send it to the socket */
  drop2socket((void *)&alert,sizeof(AlertData));
}
#endif /* DROP */


#ifdef DROP
/*
 * Function: OpenDropSocket(filename)
 * 
 * Purpose: opens unix socket "filename" for the drop feature
 *
 * Author: DG
 *
 */

void 
OpenDropSocket(char *filename)
{
  char errbuf[STRLEN];
#ifdef DEBUG_ODS
  Printf("OpenDropSocket> SocketName: %s\n",filename);
#endif

  if (filename==NULL) 
    {
      Printf("Filename missing in OpenAlertSocket!\n");
      exit(1);
    }

  if(access(filename, W_OK))
    {
      Printf("WARNING: %s file doesn't exist or isn't writable!\n", filename);
    }

  bzero((char *) &drop_addr, sizeof(drop_addr));
  bcopy((const void *)filename, (void *)drop_addr.sun_path, strlen(filename));

  drop_addr.sun_family = AF_UNIX;
  
  if((DropSocket = socket(AF_UNIX, SOCK_DGRAM, 0)) < 0)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("ODS> socket call failed: %s\n", errbuf);
      exit(1);
    }
#ifdef DEBUG_ODS
   Printf("OpenDropSocket> UnixSocket: %i\n",DropSocket);
#endif
}
#endif /* DROP */


#ifdef DROP

/*
 * Function: DropPackets(packet_ptr)
 * 
 * Purpose: calculates which packets should be dropped and
 *          writes them via DropAlertPackets to a unix socket
 *          We drop as long as the number of buffered alerts
 *          is higher than the LowWater mark
 *
 * Author: DG
 *
 */

void 
DropPackets(DGPacket *last)
{
  DGPacket *Packet;
  
#ifdef DEBUG_DROPPACKETS
  Printf("DropPackets> Begin to drop packets: %u\n",DropCounter);
#endif
  while ((RFCcounter+RSTcounter-DropCounter-WTScounter) > LowWater)
    {
      Packet=last->prev;
      DropAlert(Packet);
      last->prev=Packet->prev;
      if (Packet->pkt_ptr) free(Packet->pkt_ptr);
      free(Packet);
      DropCounter++;
    }
  Packet=last->prev;
  if (Packet) Packet->next=last;
#ifdef DEBUG_DROPPACKETS
  Printf("DropPackets> End of dropping packets: %u\n",DropCounter);
#endif
}
#endif


/*
 * Function: ReadData(filename)
 * 
 * Purpose: reads stored alerts (with payload) which were
 *          swapped out due to an error with the database
 *
 * Returns: on error -> -1
 *          else     -> number of read alert packets
 *
 * Author: DG
 *
 */

int
ReadData(char *SwapFile)
{
  AlertpktDB alertpkt;
  DGPacket *last;
  static char buf[MAXBUFFSIZE];
  static char errbuf[STRLEN];
  unsigned int size;
  int restored=0;
  int saved=0;
  
  FILE *fd;


  fd=fopen(SwapFile,"r");
  if (fd == NULL)
    {
      /* should never happen...*/
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error opening swap file %s: %s\n",SwapFile,errbuf);
      return(-1);
    }

  while(1)
    {
      /* Read alert information */
      size=sizeof(AlertpktDB);
      if (fread((void *) &alertpkt,size,1,fd) != 1) 
	{
	  /* either an error occured or we reached the end of the file */
	  if (feof(fd))
	    {
	      /* EOF */
	      fclose(fd); 
	      return(restored); 
	    }
	  else 
	    {
              fclose(fd);
	      DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("Error reading swapped alert data: %s\n",errbuf);
	      return(-1);
	    }
	}

#ifdef DEBUG_RD
      Printf("Alert (%i): %s\n",restored,alertpkt.alertmsg);
      Printf("id : %i\n",alertpkt.event.event_id);
      Printf("sid: %i\n",alertpkt.event.sig_id);
#endif
      if(!(alertpkt.val & NOPACKET_STRUCT)) 
	{
	  /* This is an alert with packet structure, not all have such */
	  /* a structure, e.g. portscans have only an alert message    */
	  size=alertpkt.pkth.caplen;
	  bzero(&buf,size);

#ifdef DEBUG_RD
	  Printf("Alert with payload: %i\n",size);
#endif
	  if (fread((void *) &buf,size,1,fd) != 1) 
	    {
	      /* This must be an error! */
	      DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("Error reading swapped data payload: %s\n",errbuf);
	      fclose(fd);
	      return(-1);
	    }

	  data->pkt_ptr = malloc(alertpkt.pkth.caplen);
	  if (data->pkt_ptr == NULL)
	    {
	      fclose(fd);
	      Printf("Malloc failed in ReadData.\n");
	      return(-1);
	    }
	  if (alertpkt.data) 
	    {
	      /* Calculate Pointer to real payload of the packet */
	      alertpkt.data += (int) data->pkt_ptr;
#ifdef DEBUG_RD
	      Printf("pkt_ptr: %p, Data: %p \n",data->pkt_ptr,alertpkt.data);
#endif
	    }
 	  bcopy(&buf, data->pkt_ptr, alertpkt.pkth.caplen);
       }

      /* read in counter of stored alerts */
      size=sizeof(saved);
      
      if (fread((void *) &saved,size,1,fd) != 1) 
	{
	  /* This must be an error! */
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Error readind control counter: %s\n",errbuf);
	  fclose(fd);
	  return(-1);
	}

      if(saved != restored)
	{
	  /* This must be an error! */
	  Printf("Error: different counter (%i) <> (%i)\n",saved,restored);
	  fclose(fd);
	  return(-1);
	}

      /* Allocate memory for the next packet */
      data->next = (DGPacket *) calloc(sizeof(DGPacket), sizeof(char));
      if (data->next == NULL)
	{
	  fclose(fd);
	  Printf("Calloc failed in ReadData.\n");
	  return(-1);
	}
      last = data; /* Pointer to actual data becomes last in next data...*/
      data->alertpkt=alertpkt;
      data=data->next;
      data->prev=last; /* last pointer in new packet shows to old data */
      restored++;
    }
}


/*
 * Function: DumpData(filename, packet_ptr)
 * 
 * Purpose: write buffered alerts (with payload if available) to a 
 *          swap file (filename). As a control statement the number
 *          of saved alert is written after each record.
 *
 * Returns: on error -> -1
 *          else     -> number of saved alert packets 
 *
 * Author: DG
 *
 */

int 
DumpData(char * SwapFile,DGPacket* data)
{
  FILE *fd;
  DGPacket *next;
  AlertpktDB alertpkt;

  static char errbuf[STRLEN];
  int counter=0;
  unsigned int size;

  /* set umask to aviod access for others and restrict group 
   * to have only read access
   */
  umask(S_IRWXO|S_IXGRP|S_IWGRP);

  /* The file should not exist. Therefore we can choose "w"
   * and truncate it to zero size if it exists.
   */

  fd=fopen(SwapFile,"w");
  if (fd==NULL)
    {
      DG_strerror_r(errno,errbuf,STRLEN);
      Printf("Error opening swap file %s: %s\n",SwapFile,errbuf);
      return(-1);
    }

  /* start loop over data */
  while (data->next != NULL)
    {
      /* Write alert data */
      alertpkt=data->alertpkt;

      /* Important: Change alertpkt.data to be a relative
       * pointer to the start of the payload, not an absolute
       * pointer. These value would be useless on a restore.
       */

      alertpkt.data = alertpkt.data - (int) data->pkt_ptr;

#ifdef DEBUG_DD
      Printf("Alert: %s\n",alertpkt.alertmsg);
      Printf("id: %i\n",alertpkt.event.event_id);
      Printf("sid: %i\n",alertpkt.event.sig_id);
#endif

      size=sizeof(AlertpktDB);
      if (fwrite(&alertpkt,size,1,fd)!=1)
	{
	  /* We could not write all data */
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Error writing alert packet to swap file %s: %s\n",
		 SwapFile,errbuf);
	  fclose(fd);
	  return(-1);
	}

      /* Do we have a payload? */
      if (!(alertpkt.val & NOPACKET_STRUCT))
	{
	  size=alertpkt.pkth.caplen;

	  if (fwrite(data->pkt_ptr,size,1,fd) != 1)
	    {
	      /* We could not write all data */
	      DG_strerror_r(errno,errbuf,STRLEN);
	      Printf("Error writing alert payload to swap file %s: %s\n",
		     SwapFile,errbuf);
	      fclose(fd);
	      return(-1);
	    }
	  /* 
	   * free the payload, since we do an exit aftwerwards
	   * it is not really necessary. But we are kindly...
	   */
	  free(data->pkt_ptr);
	}
      /* go to the next alert */
      next=data->next;
      
      /* free already saved alert data */
      free(data);

      /* switch to next alert */
      data=next;
      
      /* write actual counter to file */
      size=sizeof(counter);
      if (fwrite(&counter,size,1,fd)!=1)
	{
	  /* We could not write all data */
	  DG_strerror_r(errno,errbuf,STRLEN);
	  Printf("Error writing alert counter to swap file %s: %s\n",
		 SwapFile,errbuf);
	  fclose(fd);
	  return(-1);
	}

      /* finally increase counter */
      counter++;
    }

  /* flush all buffered data out of user space */
  fflush(fd);

  /* finally close the file */
  fclose(fd);

  /* and return the counter of stored alerts */
  return(counter);
}
